package TransDB::PDBFile;

use strict;
use vars qw(@ISA);

use TransDB::Root;
use TransDB::Utils;

@ISA = qw(TransDB::Root);

sub fn2idCode {
    my($fn) = @_;

    my $idCode;

    if($fn =~ /pdb(\S{4})\.ent/) { # standard PDB files
	$idCode = $1;
	$idCode =~ tr/[a-z]/[A-Z]/;
    }
    elsif($fn =~ /(\S{4})\.pdb/) { # pdb-biounit files (which don't have the idCode in the HEADER line... Grrrr.)
	$idCode = $1;
	$idCode =~ tr/[a-z]/[A-Z]/;
    }
    elsif($fn =~ /(\S{4})-\S+?-\S+\.pdb/) { # pdb-biounit split into idcode-assembly-model
	$idCode = $1;
	$idCode =~ tr/[a-z]/[A-Z]/;
    }
    else {
	$idCode = $fn;
    }

    return $idCode;
}

sub fn2id {
    my($fn) = @_;

    my $id;
    my $source;

    TransDB::Root::warn('deprecated');

    $id = pdb_fn2id($fn);
    if(defined($id)) {
	$source = 'pdb';
    }
    else {
	($id) = pqs_fn2id($fn);
	if(defined($id)) {
	    $source = 'pqs';
	}
	else {
	    $source = '';
	    $id = '';
	}
    }

    return($source, $id);
}

sub pdb_fn2id {
    my($fn) = @_;

    my $pdb_id;

    TransDB::Root::warn('deprecated');

    if($fn =~ /pdb(\S{4}).ent/) {
	$pdb_id = $1;
    }
    else {
	return undef;
    }
    $pdb_id =~ tr/[a-z]/[A-Z]/;

    return $pdb_id;
}

sub pqs_fn2id {
    my($fn) = @_;

    my $pqs_id;
    my $pdb_id;

    TransDB::Root::warn('deprecated');

    if($fn =~ /\A(\S+)_(\S+)\.mmol/) {
	$pdb_id = $1;
	$pqs_id = join '_', $1, $2;
    }
    elsif($fn =~ /\A(\S+)\.mmol/) {
	$pdb_id = $1;
	$pqs_id = $pdb_id;
    }
    elsif($fn =~ /\A(\S+)\.water/) {
	$pdb_id = $1;
	$pqs_id = $pdb_id . '_water';
    }
    elsif($fn =~ /\A(\S+)\.unique/) {
	$pdb_id = $1;
	$pqs_id = $pdb_id . '_unique';
    }
    elsif($fn =~ /\A(\S+)\.23fold/) {
	$pdb_id = $1;
	$pqs_id = $pdb_id . '_23fold';
    }
    elsif($fn =~ /\A(\S+)\.5fold/) {
	$pdb_id = $1;
	$pqs_id = $pdb_id . '_5fold';
    }
    else {
	return undef;
    }

    $pdb_id =~ tr/[a-z]/[A-Z]/;
    $pqs_id =~ tr/[a-z]/[A-Z]/;

    return($pdb_id, $pqs_id);
}

sub to_full {
    my($source, $id, $cid) = @_;

    my $full;

    ($cid =~ /\A[a-z]\Z/) and ($cid .= $cid);

    ($full = join('|', $id, $cid)) =~ tr/[a-z]/[A-Z]/;
    $full = join '|', $source, $full;

    return $full;
}

sub full2nochain {
    my($full) = @_;

    my $nochain;

    ($nochain = $full) =~ s/\A(\S+)\|.*/$1/;

    return $nochain;
}

sub pdb_full2fn {
    my($full, $dn) = @_;

    my $id;
    my $dn_sub;
    my $fn;

    if($full =~ /\Apdb\|(\w+)/) { # works with or without chain id
	$id = $1;
	$id =~ tr/[A-Z]/[a-z]/;
	$dn_sub = substr $id, 1, 2;
	$fn = sprintf "%s/%s/pdb%s.ent", $dn, $dn_sub, $id;
    }

    return $fn;
}

sub pqs_full2fn {
    my($full, $dn) = @_;

    my $id;
    my $dn_sub;
    my $fn;
    my $ext;

    #if($full =~ /\Apqs\|(\S+?)\|/) { # chain id required
    if($full =~ /\Apqs\|(\w+)/) { # works with or without chain id
	$id = $1;
	$id =~ tr/[A-Z]/[a-z]/;
	$dn_sub = substr $id, 1, 2;

	if($id =~ /\A(\S+)_([x\d]+)\Z/) {
	    $ext = 'mmol';
	}
	elsif($id =~ /\A(\S+)_(\S+)/) {
	    $id = $1;
	    $ext = $2;

	    $id =~ tr/[A-Z]/[a-z]/;
	    $ext =~ tr/[A-Z]/[a-z]/;
	}
	else {
	    $ext = 'mmol';
	}

	$fn = sprintf "%s/%s/%s.%s", $dn, $dn_sub, $id, $ext;
    }
    else {
	warn "Error: pqs_full2fn: do not understand '$full'.";
	return undef;
    }

    return $fn;
}

sub full2fn {
    my($full, $dn) = @_;

    my $fn;

    $fn = ($full =~ /\Apdb/) ? pdb_full2fn($full, $dn) : pqs_full2fn($full, $dn);

    return $fn;
}

sub full2fn_exists {
    my($full, @dns) = @_;

    my $dn;
    my $fn;

    if($full =~ /\Apdb/) {
	foreach $dn (@dns) {
	    $fn = pdb_full2fn($full, $dn);
	    if(-e $fn) {
		last;
	    }
	    else {
		$fn .= '.gz';
		if(-e $fn) {
		    last;
		}
		else {
		    $fn = 'UNK';
		}
	    }
	}
    }
    else {
	foreach $dn (@dns) {
	    $fn = pqs_full2fn($full, $dn);
	    if(-e $fn) {
		last;
	    }
	    else {
		$fn .= '.gz';
		if(-e $fn) {
		    last;
		}
		else {
		    $fn = 'UNK';
		}
	    }
	}
    }

    #print join("\t", $full, $fn, @dns), "\n";

    return $fn;
}

sub get_fn {
    my($idcode, $assembly, $model) = @_;

    my $dnsub;
    my $fn;

    $idcode =~ tr/[A-Z]/[a-z]/;
    $dnsub = substr($idcode, 1, 2);

    if($assembly) {
	if($model) {
	    $fn = sprintf "%s/pdb-biounit/%s/%s-%s-%s.pdb.gz", $ENV{'DS'}, $dnsub, $idcode, $assembly, $model;
	}
	else {
	    $fn = sprintf "%s/pdb-biounit/%s/%s.pdb%s.gz", $ENV{'DS'}, $dnsub, $idcode, $assembly;
	}
    }
    else {
	$fn = sprintf "%s/pdb/%s/pdb%s.ent.gz", $ENV{'DS'}, $dnsub, $idcode;
    }

    if(!(-e $fn)) {
	TransDB::PDBFile->warn("file '$fn' does not exist");
	$fn = 'UNK';
    }

    return $fn;
}

sub full2cid {
    my($full) = @_;

    my $cid;

    if($full =~ /\A.*\|([^\|]+)\Z/) {
	$cid = $1;
	(length($cid) > 1) and (($cid = substr $cid, 0, 1) =~ tr/[A-Z]/[a-z]/);
    }
    else {
	$cid = 'A';
    }

    return $cid;
}

sub revdat2date {
    my($dd, $mmm, $yy) = @_;

    my $mmm2number = {
		      'JAN' =>  1,
		      'FEB' =>  2,
		      'MAR' =>  3,
		      'APR' =>  4,
		      'MAY' =>  5,
		      'JUN' =>  6,
		      'JUL' =>  7,
		      'AUG' =>  8,
		      'SEP' =>  9,
		      'OCT' => 10,
		      'NOV' => 11,
		      'DEC' => 12,
		     };
    my $datetime;

    $datetime = sprintf(
			"%4d-%02d-%02d",
			($yy < 70) ? ($yy + 2000) : ($yy + 1900), # we're screwed if the PDB format and this code lasts past 2070.
			$mmm2number->{$mmm},
			$dd,
		       );

    return $datetime;
}

sub convert_cid {
    my($cid) = @_;

    if($cid eq '') {
	$cid = 'A';
    }
    elsif($cid =~ /\A[a-z]\Z/) { # represent lowercase as double uppercase
	$cid =~ tr/[a-z]/[A-Z]/;
	$cid .= $cid;
    }

    return $cid;
}

sub chain_type {
    my($res_names) = @_;

    my $res_name;
    my $n;
    my $type;
    my $types;

    my $chain_types = {
		       'ALA' => 'peptide',
		       'ARG' => 'peptide',
		       'ASN' => 'peptide',
		       'ASP' => 'peptide',
		       'CYS' => 'peptide',
		       'GLN' => 'peptide',
		       'GLU' => 'peptide',
		       'GLY' => 'peptide',
		       'HIS' => 'peptide',
		       'ILE' => 'peptide',
		       'LEU' => 'peptide',
		       'LEU' => 'peptide',
		       'LYS' => 'peptide',
		       'MET' => 'peptide',
		       'PHE' => 'peptide',
		       'PRO' => 'peptide',
		       'SER' => 'peptide',
		       'THR' => 'peptide',
		       'TRP' => 'peptide',
		       'TYR' => 'peptide',
		       'VAL' => 'peptide',
		       'PCA' => 'peptide',
		       'ACE' => 'peptide',
		       'FOR' => 'peptide',
		       'ASX' => 'peptide',
		       'GLX' => 'peptide',
		       'A'   => 'nucleotide',
		       'C'   => 'nucleotide',
		       'G'   => 'nucleotide',
		       'T'   => 'nucleotide',
		       'U'   => 'nucleotide',
		       'DA'  => 'nucleotide',
		       'DC'  => 'nucleotide',
		       'DG'  => 'nucleotide',
		       'DT'  => 'nucleotide',
		       'DU'  => 'nucleotide',
		      };

    $types = {};
    foreach $res_name (keys %{$res_names}) {
	$n = $res_names->{$res_name};
	defined($type = $chain_types->{$res_name}) and ($types->{$type} += $n);
    }
    $types = [sort {$types->{$b} <=> $types->{$a}} keys %{$types}];
    $type = $types->[0] ? $types->[0] : 'unknown';

    return $type;
}

sub parse_pdb {
    my($fn) = @_;

    # FIXME - currently only used for outputting res_mapping
    # and for giving chain and structure information

    my $fh;
    my @data;

    my $idCode;
    my $classification;
    my $depDate;
    my $title;
    my $str;
    my $updated;
    my $molecules;
    my $cid_str;
    my $mol_id;
    my $cid2mol;

    my $name;
    my $org;
    my $altLoc;
    my $resName;
    my $resNames;
    my $types;
    my $pdbseq_lengths;
    my $cid;
    my @cids;
    my $resSeq;
    my $iCode;
    my $p_cid;
    my $p_resSeq;
    my $p_iCode;
    my $p_resName;
    my $insertions;
    my $alternatives;
    my $positions;
    my $res_mapping;

    my $pdb;
    my $chains;

    defined($fh = TransDB::Utils::openfile($fn)) or return(undef);

    $title = '';
    $molecules = {};
    $cid2mol = {};
    @cids = ();
    $resNames = {};
    $types = {};
    $pdbseq_lengths = {};
    $p_cid = '';
    $p_resSeq = -1000000;
    $p_iCode = '';
    $p_resName = '';
    $insertions = 0;
    $alternatives = {};

    $positions = positions_init();

    $idCode = fn2idCode($fn);
    $res_mapping = [];
    $chains = {};
    while(<$fh>) {
	chomp;

	if(/^ENDMDL/) {
	    last;
	}
	elsif(/^HEADER/) {
	    @data = unpack 'a6a4a40a9a3a4', $_;
	    ($classification = $data[2]) =~ s/\A\s*(.*?)\s*\Z/$1/;
	    ($data[3] =~ /\A(.{2})-(.{3})-(.{2})/) and ($depDate = revdat2date($1, $2, $3));

	    $idCode = ($data[5] and ($data[5] ne 'XXXX')) ? $data[5] : fn2idCode($fn); # pdb-biounit has XXXX in HEADER line, instead of the idCode...
	}
	elsif(/^TITLE.{5}(.{1,60})/) {
	    ($str = $1) =~ s/\s+\Z//;
	    $title .= $str;
	}
	elsif(/^REVDAT.{7}(.{2})-(.{3})-(.{2})/ and !$updated) {
	    $updated = revdat2date($1, $2, $3);
	}
	elsif(/^COMPND.*MOL_ID:\s*(\d+)/) {
	    $mol_id = $1;
	    $molecules->{$mol_id} = {'name' => '', 'chains' => [], 'org' => ''};
	}
	elsif(/^COMPND\s*\S+\s*MOLECULE:\s*(.*?)\s*\Z/) {
	    $molecules->{$mol_id}->{'name'} .= $1;
	}
	elsif(/^COMPND\s*\S+\s*CHAIN:\s*(.*)/) {
	    ($cid_str = $1) =~ s/;?\s*\Z//;

	    foreach $cid (split(/\s*,\s*/, $cid_str)) {
		(length($cid) > 1) and next; # some people abuse the CHAIN info, eg. 3e6p has 'CHAIN: UNP RESIDUES 206-363;'

		$cid = convert_cid($cid);
		push @{$molecules->{$mol_id}->{'chains'}}, $cid;
		$cid2mol->{$cid} = $mol_id;
	    }
	}
	elsif(/^SOURCE.*MOL_ID:\s*(\d+)/) {
	    $mol_id = $1;
	}
	elsif(/^SOURCE.*ORGANISM_SCIENTIFIC:\s*(.*?)\s*\Z/) {
	    $molecules->{$mol_id}->{'org'} .= $1;
	}
	elsif(/^ATOM/ or /^HETATM.{7}CA /) {
	    # Could include CAs in HETATMS with 'elsif(/^ATOM/ or /^HETATM.{7}CA /)'.
	    # CAs in HETATMs can also indicate an amino acid, from MSE etc.

	    @data = unpack 'a6a5a5aa3a2a4a', $_;
	    ($name    = $data[2]) =~ s/\s+//g;
	    ($altLoc  = $data[3]) =~ s/\s+//g;
	    ($resName = $data[4]) =~ s/\s+//g;
	    ($cid     = $data[5]) =~ s/\s+//g;
	    ($resSeq  = $data[6]) =~ s/\s+//g;
	    ($iCode   = $data[7]) =~ s/\s+//g;

	    ($iCode eq '') and ($iCode = '_'); # STAMP uses underscores to indicate spaces. Easiest for me to do the same.

	    $cid = convert_cid($cid);
	    $mol_id = $cid2mol->{$cid};

	    if($cid ne $p_cid) {
		if($p_resSeq > -1000000) {
		    $pdbseq_lengths->{$p_cid} = $positions->{'pdbseq'};
		    $types->{$p_cid} = chain_type($resNames);

		    push @{$res_mapping}, [
					   $idCode,
					   $p_cid,
					   $p_resSeq,
					   $p_iCode,
					   $p_resName,
					   $positions->{'ca_inc_ins'},
					   $positions->{'res_inc_ins'},
					   $positions->{'ca_no_ins'},
					   $positions->{'res_no_ins'},
					   $positions->{'pdbseq'},
					  ];
		}

		push @cids, $cid;

		$p_resSeq = -1000000;
		$p_iCode = '';
		$p_resName = '';
		$insertions = 0;
		$alternatives = {};
		$resNames = {};

		$positions = positions_init();

		$chains->{$cid} = {'cid' => $cid}; # other info (species etc.) set later
	    }

	    if($resSeq eq $p_resSeq) {
		if($iCode ne $p_iCode) {
		    if($p_resSeq) {
			push @{$res_mapping}, [
					       $idCode,
					       $cid,
					       $p_resSeq,
					       $p_iCode,
					       $p_resName,
					       $positions->{'ca_inc_ins'},
					       $positions->{'res_inc_ins'},
					       $positions->{'ca_no_ins'},
					       $positions->{'res_no_ins'},
					       $positions->{'pdbseq'},
					      ];
		    }

		    $resNames->{$p_resName}++;

		    $positions->{'res_inc_ins'}++;
		    $insertions++;
		    $alternatives = {};
		}
	    }
	    else {
		if($p_resSeq > -1000000) {
		    push @{$res_mapping}, [
					   $idCode,
					   $cid,
					   $p_resSeq,
					   $p_iCode,
					   $p_resName,
					   $positions->{'ca_inc_ins'},
					   $positions->{'res_inc_ins'},
					   $positions->{'ca_no_ins'},
					   $positions->{'res_no_ins'},
					   $positions->{'pdbseq'},
					  ];

		    $resNames->{$p_resName}++;
		}

		$positions->{'res_no_ins'}++;
		$positions->{'res_inc_ins'}++;

		$insertions = 0;
		$alternatives = {};
	    }

	    if($name eq 'CA') {
		$alternatives->{$altLoc}++;

		# only do this for the first of any alternative locations for the CA
		if(keys %{$alternatives} <= 1) {
		    $positions->{'ca_inc_ins'}++;
		    $insertions or $positions->{'ca_no_ins'}++;
		}

		if(($altLoc eq '') or ($altLoc eq 'A') or ($altLoc eq '1') or ($altLoc eq 'L') or ($altLoc eq 'O')) {
		    $positions->{'pdbseq'}++;
		}
	    }

	    $p_cid = $cid;
	    $p_resSeq = $resSeq;
	    $p_resName = $resName;
	    $p_iCode = $iCode;
	}
    }
    if($p_resSeq > -1000000) {
	$pdbseq_lengths->{$cid} = $positions->{'pdbseq'};
	$types->{$cid} = chain_type($resNames);
	push @{$res_mapping}, [
			       $idCode,
			       $cid,
			       $p_resSeq,
			       $p_iCode,
			       $p_resName,
			       $positions->{'ca_inc_ins'},
			       $positions->{'res_inc_ins'},
			       $positions->{'ca_no_ins'},
			       $positions->{'res_no_ins'},
			       $positions->{'pdbseq'},
			      ];
    }
    close($fh);

    # FIXME - $pdb and $chains should be proper objects

    $pdb = {'idCode' => $idCode, 'title' => $title, 'depdate' => $depDate, 'updated' => $updated};

    #print join("\t", $idCode, sort keys %{$cid2mol}), "\n";

    foreach $cid (sort keys %{$cid2mol}) {
	$mol_id = $cid2mol->{$cid};

	$molecules->{$mol_id}->{'org'} =~ s/;\s*\Z//;
	$molecules->{$mol_id}->{'name'} =~ s/;\s*\Z//;

	$chains->{$cid} = {
			   'chemical_type' => $types->{$cid} ? $types->{$cid} : 'unknown',
			   'cid'           => $cid,
			   'organism'      => $molecules->{$mol_id}->{'org'},
			   'name'          => $molecules->{$mol_id}->{'name'},
			  };
    }

    return($res_mapping, $pdb, $chains);
}

sub positions_init {
    my $positions;

    $positions = {
		    'ca_no_ins'   => 0,
		    'ca_inc_ins'  => 0,
		    'res_no_ins'  => 0,
		    'res_inc_ins' => 0,
		    'pdbseq'      => 0,
		   };

    return $positions;
}

sub aa3to1 {
    my($aa3) = @_;

    my $aa1;
    my $aas = {
	       'ALA' => 'A',
	       'ARG' => 'R',
	       'ASN' => 'N',
	       'ASP' => 'D',
	       'CYS' => 'C',
	       'GLN' => 'Q',
	       'GLU' => 'E',
	       'GLY' => 'G',
	       'HIS' => 'H',
	       'ILE' => 'I',
	       'LEU' => 'L',
	       'LYS' => 'K',
	       'MET' => 'M',
	       'PHE' => 'F',
	       'PRO' => 'P',
	       'SER' => 'S',
	       'THR' => 'T',
	       'TRP' => 'W',
	       'TYR' => 'Y',
	       'VAL' => 'V',
	       'ASX' => 'B',
	       'GLX' => 'Z',
	       'PCA' => 'E',
	       'ACE' => 'X',
	       'FOR' => 'X',
	      };

    $aa1 = $aas->{$aa3} ? $aas->{$aa3} : 'X';

    return $aa1;
}

1;
