package Web;

use strict;

use CGI;

sub ssi {
    my($fn) = @_;

    my @lines;
    my $fh;

    # Can't see how to use SSI in CGI scripts, so am reading the included file myself
    @lines = ();
    if($ENV{'DOCUMENT_ROOT'}) {
	$fn = join '/', $ENV{'DOCUMENT_ROOT'}, $fn;
	if(-e $fn) {
	    if(!open($fh, $fn)) {
		warn "Error: Web::ssi: cannot open '$fn' for reading.";
	    }
	    else {
		@lines = <$fh>;
		close($fh);
	    }
	}
    }

    return(@lines);
}

sub head {
    my($cgi, $title, $name, $fh, $head) = @_;

    defined($fh) or ($fh = \*STDOUT);

    my @header;

    @header = ssi('header.shtml');

    print(
	  $fh
	  ($fh == \*STDOUT) ? $cgi->header() : '', # don't need $cgi->header() if printing to a file
	  $cgi->start_html(
			   '-title'  => $title,
			   '-style'  => {'src' => '/default.css'},
			   '-script' => [ # FIXME - only include this when needed
					 {'-language' => 'JAVASCRIPT', '-src' => '/js/toggle.js'},
					 {'-language' => 'JAVASCRIPT', '-src' => '/jmol/Jmol.js'},
					],
			   '-head'   => $head,
			  ),
	  "\n",
	  "<div id=\"wrapper\">\n",
	  @header,
	  "<div id=\"page\">\n",
	 );
}

sub tail {
    my($cgi, $name, $fh) = @_;

    defined($fh) or ($fh = \*STDOUT);

    print(
	  $fh
	  "</div> <!-- end of page div -->\n",
	  "</div> <!-- end of wrapper div -->\n",
	  $cgi->end_html(),
	  "\n",
	  );
}

sub menu {
    my($cgi, $active) = @_;

    my $menu;

    # FIXME - there's probably a better way of highlighting the active item...

    $menu = join(' | ',
		 sprintf("<a %shref=\"/cgi-bin/tools/pdbseq.pl\">PDBseq</a>", ($active eq 'pdbseq') ? 'class="active" ' : ''),
		 sprintf("<a %shref=\"/cgi-bin/tools/interprets.pl\">InterPreTS</a>", ($active eq 'interprets') ? 'class="active" ' : ''),
		 sprintf("<a %shref=\"/cgi-bin/tools/stamp.pl\">Superposition</a>", ($active eq 'stamp') ? 'class="active" ' : ''),
		 sprintf("<a %shref=\"/cgi-bin/tools/jmol.pl\">Jmol</a>", ($active eq 'jmol') ? 'class="active" ' : ''),
		);

    return $menu;
}

sub error_string {
    my($cgi, $error) = @_;

    my $err_str;

    $err_str = '';
    if($error) {
	$err_str = join('',
			"<div id=\"error\" class=\"error\">\n",
			$cgi->h2('Error'),
			"<ul>\n",
			"<li>$error</li>",
			"<li>Please <a href=\"mailto:betts\@embl.de\">ask me</a> for help or if you think this is a bug.</li>",
			"</ul>\n",
			"</div>\n",
		       );
    }

    return $err_str;
}

sub error {
    my($cgi, $error) = @_;

    # FIXME - make errors appear just under the header

    $error and print(error_string($cgi, $error));
}

sub job_id {
    my($cgi, $tmp_dn) = @_;

    my $i = 0;
    my $time;
    my $id;
    my $fn;
    my $fh;

    $time = time;
    $id = join '_', $time, $i;
    $fn = join '', $tmp_dn, $id, '.job';
    while(-e $fn) {
	$i++;
	$id = join '_', $time, $i;
	$fn = join '', $tmp_dn, $id, '.job';
    }
    open($fh, ">$fn") or error($cgi, "Cannot open '$fn' file for writing.");
    close($fh);

    return $id;
}

sub mysystem {
    # http://www.perlmonks.org/?node_id=197500
    #
    # Use local signal handler so global handler
    # does not result in bad values in $? and $!

    local $SIG{CHLD} = '';
    return(system(@_));
}

1;
